import csv
from saga.file_types.binary_file import parse_binary_file, write_binary_file
from saga.file_types.text_file import parse_text_file, write_text_file
from saga.file_types.csv_file import parse_csv_file, write_csv_file
from saga.file_types.excel_file import parse_excel_file, write_excel_file

def is_csv(file_path):
    return file_path.endswith(".csv")

TEXT_ENDINGS = {
    "txt", # just text
    "py", # python
    "c", # c source
    "cc", # c++ source 
    "java", # java source 
    "js",
    "php"
}

def is_text(file_path):
    try:
        ending = file_path.split(".")[-1]
        return ending in TEXT_ENDINGS
    except:
        return False

def is_excel(file_path):
    return file_path.endswith(".xlsx")


def parse_file(file_id, file_name, file_path):
    if is_text(file_path):
        return parse_text_file(file_id, file_name, file_path)
    elif is_excel(file_path):
        return parse_excel_file(file_id, file_name, file_path)
    elif is_csv(file_path):
        return parse_csv_file(file_id, file_name, file_path)
    else:
        return parse_binary_file(file_id, file_name, file_path)


def write_file(file):
    if file.file_type == "csv":
        write_csv_file(file)
    elif file.file_type == "text":
        write_text_file(file)
    elif file.file_type == "binary":
        write_binary_file(file)
    elif file.file_type == "excel":
        write_excel_file(file)
    else:
        print("Error: file type {} does not exist".format(file.file_type))


from saga.base_file.mixed_data_type.lcs import lcs_with_sim

PRIMITIVE = (int, float, bool, str)

def paths(A, B):
    # returns a triple (removed_paths, changed_paths, inserted_paths)
    if type(A) in PRIMITIVE or type(B) in PRIMITIVE:
        assert type(A) in PRIMITIVE and type(B) in PRIMITIVE
        return [], [], [] if A == B else [], [[]], []

def changed_paths(A, B):
    return _changed_paths(A, B, [])

def _changed_paths(A, B, base_path):
    if type(A) in PRIMITIVE or type(B) in PRIMITIVE:
        assert type(A) in PRIMITIVE and type(B) in PRIMITIVE
        return [] if A == B else [base_path]

    # All variables must have the same type!
    assert type(A) == type(B)

    changed_paths = []
    
    if type(A) == dict:
        changed_keys = dict_changed_paths(A, B)

        # we only report changes on the lowest level
        for key in changed_keys:
            changed_paths.extend(_changed_paths(A[key], B[key], base_path + [key]))
        
        return changed_paths

    elif type(A) == list:
        lcs = lcs_with_sim(A, B)
        count = 0
        for path_a, path_b, sim in lcs:
            if sim < 1:
                count += 1
                changed_paths.extend(_changed_paths(A[path_a[0]], B[path_b[0]], base_path + path_a))
        if count == len(A) and count > 0 and len(base_path) > 0:
            # the whole list was changed, and so we just say this list was changed
            return [base_path]

        return changed_paths
    else:
        raise ValueError("Error: unknown type given: {}".format(type(A)))
    
def removed_paths(A, B):
    return _removed_paths(A, B, [])

def _removed_paths(A, B, base_path):
    if type(A) in PRIMITIVE or type(B) in PRIMITIVE:
        return []

    # All variables must have the same type!
    assert type(A) == type(B)

    if type(A) == dict:
        removed = []
        removed_keys = dict_removed_paths(A, B)
        # if the whole dictonary has been changed we report that
        if len(removed_keys) == len(A) and len(A) > 0:
            return [base_path]
        # otherwise
        removed.extend([base_path + [key] for key in removed_keys])

        # we also have to recurse on all the changed paths
        changed_keys = dict_changed_paths(A, B)
        for key in changed_keys:
            removed.extend(_removed_paths(A[key], B[key], base_path + [key]))
        
        # we don't recurse, as it doesn't matter what happened below
        return removed

    elif type(A) == list:
        removed_paths = []
        lcs = lcs_with_sim(A, B)
        # nothing was matched, so we removed all the objects
        # but if nothing was matched as it was empty, then it wasn't deleted
        if len(lcs) == 0 and len(A) > 0 and len(base_path) > 0:
            return [base_path]

        for idx in range(len(A)):
            matched = False
            for [idx_a], [idx_b], sim in lcs:
                if idx == idx_a:
                    matched = True
                    break
            if not matched:
                removed_paths.append(base_path + [idx])
            elif matched and sim < 1:
                # if the subtype of the list is not primitive, we have to recurse here also
                if type(A[idx]) not in PRIMITIVE:
                    removed_paths.extend(_removed_paths(A[idx], B[idx_b], base_path + [idx]))

                # we have to recurse

        return removed_paths
    else:
        raise ValueError("Error: unknown type given: {}".format(type(A)))

def inserted_paths(A, B):
    return removed_paths(B, A)

def dict_removed_paths(dict_a, dict_b):
    return set(dict_a.keys()).difference(dict_b.keys())

def dict_inserted_paths(dict_a, dict_b):
    return dict_removed_paths(dict_b, dict_a)

def dict_changed_paths(dict_a, dict_b):
    return {k for k in dict_a if k in dict_b and dict_a[k] != dict_b[k]}





from saga.base_file.File import File
import csv

def parse_csv_file(file_id, file_name, file_path):
    file_contents = []
    with open(file_path) as csvfile:
        csvreader = csv.reader(csvfile)
        for row in csvreader:
            file_contents.append(row)

    return File(file_id, "csv", file_path, file_name, file_contents)

def write_csv_file(file):
    lines = [",".join(row) for row in file.file_contents.mixed_data_type]
    f = open(file.file_path, "w+")
    f.write("\n".join(lines))
    f.close()




from saga.base_file.mixed_data_type.merge_utils import merge_rec, equal_objs
from saga.base_file.mixed_data_type.diff_utils import inserted_paths, removed_paths, changed_paths

def value_at_path(obj, path):
    if len(path) == 0:
        return obj
    else:
        return value_at_path(obj[path[0]], path[1:])

def dict_same_paths(dict_a, dict_b):
    return {k for k in dict_a if k in dict_b and dict_a[k] == dict_b[k]}

def dict_removed_paths(dict_a, dict_b):
    return set(dict_a.keys()).difference(dict_b.keys())

def dict_inserted_paths(dict_a, dict_b):
    return removed_paths(dict_b, dict_a)

def dict_changed_paths(dict_a, dict_b):
    return {k for k in dict_a if k in dict_b and dict_a[k] != dict_b[k]}

class MixedDataType(object):

    def __init__(self, mixed_data_type):
        self.mixed_data_type = mixed_data_type

    def get_operations(self, mixed_data_type_obj):
        A =  self.mixed_data_type
        B = mixed_data_type_obj.mixed_data_type

        removed = removed_paths(A, B)
        changed = changed_paths(A, B)
        inserted = inserted_paths(A, B)

        return {"removed": removed, "changed": changed, "inserted": inserted}

    def merge(self, a_mdt, b_mdt):
        O = self.mixed_data_type
        A = a_mdt.mixed_data_type
        B = b_mdt.mixed_data_type

        merged, err = merge_rec(O, A, B)
        return MixedDataType(merged)

    def __eq__(self, value):
        if not isinstance(value, MixedDataType):
            return False
        return equal_objs(self.mixed_data_type, value.mixed_data_type)

PRIMITIVE = (int, float, bool, str, type(None))

    


#!/usr/bin/env python3
import argparse
import os
from saga.Repository import Repository

def main():
    # create the top-level saga parser
    parser = argparse.ArgumentParser(prog='saga')
    subparsers = parser.add_subparsers(help='sub-command help')
    
    # create the parser for the "init" command
    parser_init = subparsers.add_parser('init', help='creates a new saga project in the current directory')
    parser_init.set_defaults(func=init)

    # create the parser for the "add" command
    parser_add = subparsers.add_parser('add', help='adds a path to the current index')
    parser_add.add_argument('path', type=str, help='the path to add to the index')
    parser_add.set_defaults(func=add)

    # create the parser for the "commit" command
    parser_commit = subparsers.add_parser('commit', help='adds a path to the current index')
    parser_commit.add_argument('-m', type=str, help='commit string')
    parser_commit.set_defaults(func=commit)

    # create the parser for the "status" command
    parser_status = subparsers.add_parser('status', help='displays info about the current working index')
    parser_status.set_defaults(func=status)

    # create the parser for the "log" command
    parser_log = subparsers.add_parser('log', help='displays most recent commits')
    parser_log.set_defaults(func=log)

    # create the parser for the "diff" command
    parser_diff = subparsers.add_parser('diff', help='check the diff of the current index')
    parser_diff.set_defaults(func=diff)

    # create the parser for the "branch" command
    parser_branch = subparsers.add_parser('branch', help='commands for managing branches')
    parser_branch.add_argument('b', nargs="?", type=str, help='new branch to create')
    parser_branch.set_defaults(func=branch)

    # create the parser for the "checkout" command
    parser_checkout = subparsers.add_parser('checkout', help='switches the head branch')
    parser_checkout.add_argument('-b', help='flag to create a new branch', action='store_true')
    parser_checkout.add_argument('branch', type=str, help='branch name')
    parser_checkout.set_defaults(func=checkout)

    # create the parser for the "merge" command
    parser_merge = subparsers.add_parser('merge', help='merges the head branch with some other branch')
    parser_merge.add_argument('branch', type=str, help='name of branch to merge')
    parser_merge.set_defaults(func=merge)

    # create the parser for the "push" command
    parser_push = subparsers.add_parser('push', help='uploads the local repository to the remote repository')
    parser_push.set_defaults(func=push)

    # create the parser for the "pull" command
    parser_pull = subparsers.add_parser('pull', help='downloads the local repository from the remote repository')
    parser_pull.set_defaults(func=pull)

    args = parser.parse_args()
    parser.parse_args()
    args.func(args)

def init(args):
    saga_repo = get_saga_repo_maybe()
    if saga_repo is not None:
        print("Error: saga project already exists at {}".format(saga_repo.base_directory))
    else:
        repository = Repository.init(os.getcwd())
        repository.write()
        print("Initialized empty saga repository in {}".format(repository.saga_directory))

def add(args):
    saga_repo = get_saga_repo()
    saga_repo.add(args.path)
    saga_repo.write()

def commit(args):
    saga_repo = get_saga_repo()
    if args.m is None:
        commit_message = input("Please enter a commit message: ")
        saga_repo.commit(commit_message)
    else:
        saga_repo.commit(args.m)
    saga_repo.write()

def status(args):
    saga_repo = get_saga_repo()
    saga_repo.status()

def log(args):
    saga_repo = get_saga_repo()
    saga_repo.log()

def diff(args):
    get_saga_repo().diff()

def branch(args):
    saga_repo = get_saga_repo()
    if args.b is not None:
        saga_repo.create_branch(args.b)
    saga_repo.branch()
    saga_repo.write()

def checkout(args):
    saga_repo = get_saga_repo()
    if args.b:
        saga_repo.create_branch(args.branch)
    saga_repo.switch_to_branch(args.branch)
    saga_repo.write()

def merge(args):
    saga_repo = get_saga_repo()
    saga_repo.merge(args.branch)
    saga_repo.write()

def push(args):
    print("PUSHING")
    saga_repo = get_saga_repo()
    saga_repo.push()

def pull(args):
    saga_repo = get_saga_repo()
    saga_repo.pull()
    saga_repo = get_saga_repo()
    saga_repo.restore_state_to_head()

# Helper functions

def get_saga_repo():
    saga_repo = get_saga_repo_maybe()
    if saga_repo is None:
        print("Error: command cannot run as no saga repo exists")
        exit(1)
    else:
        return saga_repo

def get_saga_repo_maybe():
    path = os.getcwd()
    while path != "/":
        # check if there is a
        if os.path.isdir(path + "/.saga"):
            return Repository.read(path) 
        path = os.path.dirname(path)
    return None

if __name__ == "__main__":
    main()



import os
import subprocess
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

commit_message = subprocess.check_output(['git', "log", "-1", "--pretty=%B"]).decode("utf-8")

# the last element must be a version num er
version = str(commit_message.split(" ")[-1]).strip()
print("COMMIT MESSAGE {}".format(commit_message))
print("Uploading version {}".format(version))

setuptools.setup(
    name="saga-vcs",
    version=version,
    author="saga",
    author_email="narush@wharton.upenn.edu.com",
    description="saga is a version control CLI that handles many file formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/saga-vcs/saga",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=[
        'XlsxWriter>=1.2',
        'beautifulsoup4>=4.8',
        'html5lib>=1.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
     entry_points={
        'console_scripts': [
            'saga = saga.main:main'
        ]
    },
    python_requires='>=3.6',
)


from saga.base_file.mixed_data_type.MixedDataType import MixedDataType

class File(object):

    def __init__(self, file_id, file_type, file_path, file_name, file_contents):
        """
        Create a new representation of a file:
            - file_id, file_type are static and can never change
            - file_name, file_contents can
        """
        self.file_id = file_id
        self.file_type = file_type
        self.file_path = file_path
        self.file_name = file_name
        self.file_contents = MixedDataType(file_contents)

    def get_operations(self, other_file):
        """
        Get the operations between this file and the new version of the file, other_file
        """
        if self.file_id != other_file.file_id:
            raise ValueError("Can only get operations between different versions of the same file")

        operations = {"file": []}
        if self.file_name != other_file.file_name:
            operations["file"].append("Change file name from {} to {}".format(self.file_name, other_file.file_name))

        # now we delegate the operations to the file_contents
        operations.update(self.file_contents.get_operations(other_file.file_contents))
        return operations

    # this should be called on the origin file
    def merge(self, file_a, file_b, file_path=None):
        merge_res = self.file_contents.merge(file_a.file_contents, file_b.file_contents)
        if merge_res is None:
            return None
        return File(self.file_id, self.file_type, file_path, self.file_name, merge_res.mixed_data_type)
